@@ -63,7 +63,7 @@ class UTXOData(NamedTuple):
     utxo_hash: str
     utxo_ix: str
     amount: int
-    address: Optional[str] = None
+    address: str
     coin: str = DEFAULT_COIN
 
 
@@ -79,12 +79,23 @@ OptionalTxOuts = Union[List[TxOut], Tuple[()]]
 OptionalUTXOData = Union[List[UTXOData], Tuple[()]]
 
 
+class ScriptFiles(NamedTuple):
+    txin_scripts: OptionalFiles = ()
+    minting_scripts: OptionalFiles = ()
+    certificate_scripts: OptionalFiles = ()
+    withdrawal_scripts: OptionalFiles = ()
+    auxiliary_scripts: OptionalFiles = ()
+
+
+OptionalScriptFiles = Union[ScriptFiles, Tuple[()]]
+
+
 class TxFiles(NamedTuple):
     certificate_files: OptionalFiles = ()
     proposal_files: OptionalFiles = ()
     metadata_json_files: OptionalFiles = ()
     metadata_cbor_files: OptionalFiles = ()
-    script_files: OptionalFiles = ()
+    script_files: OptionalScriptFiles = ()
     signing_key_files: OptionalFiles = ()
 
 
@@ -178,7 +189,7 @@ class ClusterLib:
     Attributes:
         state_dir: A directory with cluster state files (keys, config files, logs, ...).
         protocol: A cluster protocol - full cardano mode by default.
-        tx_era: An era used for transactions, by default same as `era`.
+        tx_era: An era used for transactions, by default same as network Era.
         slots_offset: Difference in slots between cluster's start era and current era
             (e.g. Byron->Mary)
     """
@@ -431,13 +442,7 @@ class ClusterLib:
         utxo = []
         for utxo_rec, utxo_data in utxo_dict.items():
             utxo_hash, utxo_ix = utxo_rec.split("#")
-            # TODO: workaround for https://github.com/input-output-hk/cardano-node/issues/2460
-            if "address" not in utxo_data:
-                addr = next(iter(utxo_data))  # first key
-                addr_data = next(iter(utxo_data.values()))  # first value
-            else:
-                addr = utxo_data["address"]
-                addr_data = utxo_data["value"]
+            addr_data = utxo_data["value"]
             for policyid, coin_data in addr_data.items():
                 if policyid == DEFAULT_COIN:
                     utxo.append(
@@ -445,7 +450,7 @@ class ClusterLib:
                             utxo_hash=utxo_hash,
                             utxo_ix=utxo_ix,
                             amount=coin_data,
-                            address=addr,
+                            address=address,
                             coin=DEFAULT_COIN,
                         )
                     )
@@ -456,7 +461,7 @@ class ClusterLib:
                             utxo_hash=utxo_hash,
                             utxo_ix=utxo_ix,
                             amount=amount,
-                            address=addr,
+                            address=address,
                             coin=f"{policyid}.{asset_name}" if asset_name else policyid,
                         )
                     )
@@ -1243,8 +1248,8 @@ class ClusterLib:
         """Return last block KES period."""
         return int(self.get_slot_no() // self.slots_per_kes_period)
 
-    def get_txid(self, tx_body_file: FileType) -> str:
-        """Get the transaction identifier trom transaction body.
+    def get_txid_body(self, tx_body_file: FileType) -> str:
+        """Get the transaction identifier from transaction body (JSON TxBody).
 
         Args:
             tx_body_file: A path to the transaction body file.
@@ -1258,6 +1263,21 @@ class ClusterLib:
             .decode("ascii")
         )
 
+    def get_txid_signed(self, tx_file: FileType) -> str:
+        """Get the transaction identifier from signed transaction (JSON Tx).
+
+        Args:
+            tx_file: A path to the signed transaction file.
+
+        Returns:
+            str: A transaction ID.
+        """
+        return (
+            self.cli(["transaction", "txid", "--tx-file", str(tx_file)])
+            .stdout.rstrip()
+            .decode("ascii")
+        )
+
     def get_tx_deposit(self, tx_files: TxFiles) -> int:
         """Get deposit amount for a transaction (based on certificates used for the TX).
 
@@ -1398,7 +1418,7 @@ class ClusterLib:
         deposit: Optional[int],
         withdrawals: OptionalTxOuts,
     ) -> List[TxOut]:
-        """Ballance the transaction by adding change output for each coin."""
+        """Balance the transaction by adding change output for each coin."""
         txouts_result: List[TxOut] = []
 
         # iterate over coins both in txins and txouts
@@ -1462,7 +1482,7 @@ class ClusterLib:
         Args:
             src_address: An address used for fee and inputs (if inputs not specified by `txins`).
             tx_files: A `TxFiles` tuple containing files needed for the transaction.
-            txins: An interable of `UTXOData`, specifying input UTxOs (optional).
+            txins: An iterable of `UTXOData`, specifying input UTxOs (optional).
             txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
             fee: A fee amount (optional).
             deposit: A deposit amount needed by the transaction (optional).
@@ -1578,7 +1598,7 @@ class ClusterLib:
 
         Args:
             out_file: An output file.
-            txins: An interable of `UTXOData`, specifying input UTxOs.
+            txins: An iterable of `UTXOData`, specifying input UTxOs.
             txouts: A list (iterable) of `TxOuts`, specifying transaction outputs.
             tx_files: A `TxFiles` tuple containing files needed for the transaction.
             fee: A fee amount.
@@ -1631,6 +1651,22 @@ class ClusterLib:
         mint_records = [f"{m.amount} {m.coin}" for m in mint]
         mint_args = ["--mint", "+".join(mint_records)] if mint_records else []
 
+        script_args = []
+        if tx_files.script_files:
+            script_args = [
+                *self._prepend_flag("--txin-script-file", tx_files.script_files.txin_scripts),
+                *self._prepend_flag("--minting-script-file", tx_files.script_files.minting_scripts),
+                *self._prepend_flag(
+                    "--certificate-script-file", tx_files.script_files.certificate_scripts
+                ),
+                *self._prepend_flag(
+                    "--withdrawal-script-file", tx_files.script_files.withdrawal_scripts
+                ),
+                *self._prepend_flag(
+                    "--auxiliary-script-file", tx_files.script_files.auxiliary_scripts
+                ),
+            ]
+
         self.cli(
             [
                 "transaction",
@@ -1645,10 +1681,10 @@ class ClusterLib:
                 *self._prepend_flag("--update-proposal-file", tx_files.proposal_files),
                 *self._prepend_flag("--metadata-json-file", tx_files.metadata_json_files),
                 *self._prepend_flag("--metadata-cbor-file", tx_files.metadata_cbor_files),
-                *self._prepend_flag("--script-file", tx_files.script_files),
                 *self._prepend_flag("--withdrawal", withdrawals_combined),
                 *bound_args,
                 *mint_args,
+                *script_args,
                 *self.tx_era_arg,
             ]
         )
@@ -1685,7 +1721,7 @@ class ClusterLib:
         Args:
             src_address: An address used for fee and inputs (if inputs not specified by `txins`).
             tx_name: A name of the transaction.
-            txins: An interable of `UTXOData`, specifying input UTxOs (optional).
+            txins: An iterable of `UTXOData`, specifying input UTxOs (optional).
             txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
             tx_files: A `TxFiles` tuple containing files needed for the transaction (optional).
             fee: A fee amount (optional).
@@ -1803,7 +1839,7 @@ class ClusterLib:
             src_address: An address used for fee and inputs (if inputs not specified by `txins`).
             tx_name: A name of the transaction.
             dst_addresses: A list of destination addresses (optional)
-            txins: An interable of `UTXOData`, specifying input UTxOs (optional).
+            txins: An iterable of `UTXOData`, specifying input UTxOs (optional).
             txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
             tx_files: A `TxFiles` tuple containing files needed for the transaction (optional).
             ttl: A last block when the transaction is still valid
@@ -1859,7 +1895,6 @@ class ClusterLib:
         tx_body_file: FileType,
         signing_key_files: OptionalFiles,
         tx_name: str,
-        script_files: OptionalFiles = (),
         destination_dir: FileType = ".",
     ) -> Path:
         """Sign a transaction.
@@ -1868,7 +1903,6 @@ class ClusterLib:
             tx_body_file: A path to file with transaction body.
             signing_key_files: A list of paths to signing key files.
             tx_name: A name of the transaction.
-            script_files: A list of paths to script files (optional).
             destination_dir: A path to directory for storing artifacts (optional).
 
         Returns:
@@ -1887,7 +1921,6 @@ class ClusterLib:
                 str(out_file),
                 *self.magic_args,
                 *self._prepend_flag("--signing-key-file", signing_key_files),
-                *self._prepend_flag("--script-file", script_files),
             ]
         )
 
@@ -1899,7 +1932,6 @@ class ClusterLib:
         tx_body_file: FileType,
         witness_name: str,
         signing_key_files: OptionalFiles = (),
-        script_file: Optional[FileType] = None,
         destination_dir: FileType = ".",
     ) -> Path:
         """Create a transaction witness.
@@ -1908,7 +1940,6 @@ class ClusterLib:
             tx_body_file: A path to file with transaction body.
             witness_name: A name of the transaction witness.
             signing_key_files: A list of paths to signing key files (optional).
-            script_file: A path to script file (optional).
             destination_dir: A path to directory for storing artifacts (optional).
 
         Returns:
@@ -1917,10 +1948,6 @@ class ClusterLib:
         destination_dir = Path(destination_dir).expanduser()
         out_file = destination_dir / f"{witness_name}_tx.witness"
 
-        cli_args = []
-        if script_file:
-            cli_args = ["--script-file", str(script_file)]
-
         self.cli(
             [
                 "transaction",
@@ -1931,7 +1958,6 @@ class ClusterLib:
                 str(out_file),
                 *self.magic_args,
                 *self._prepend_flag("--signing-key-file", signing_key_files),
-                *cli_args,
             ]
         )
 
@@ -1974,8 +2000,8 @@ class ClusterLib:
         self._check_outfiles(out_file)
         return out_file
 
-    def submit_tx(self, tx_file: FileType) -> None:
-        """Submit a transaction.
+    def submit_tx_bare(self, tx_file: FileType) -> None:
+        """Submit a transaction, don't do any verification that it made it to the chain.
 
         Args:
             tx_file: A path to signed transaction file.
@@ -1991,6 +2017,45 @@ class ClusterLib:
             ]
         )
 
+    def submit_tx(self, tx_file: FileType, txins: List[UTXOData], wait_blocks: int = 2) -> None:
+        """Submit a transaction, resubmit if the transaction didn't make it to the chain.
+
+        Args:
+            tx_file: A path to signed transaction file.
+            txins: An iterable of `UTXOData`, specifying input UTxOs.
+            wait_blocks: A number of new blocks to wait for (default = 2).
+        """
+        txid = ""
+        for r in range(3):
+            if r != 0:
+                txid = txid or self.get_txid_signed(tx_file)
+                LOGGER.info(f"Resubmitting transaction '{txid}' (from '{tx_file}').")
+
+            self.submit_tx_bare(tx_file)
+            self.wait_for_new_block(wait_blocks)
+
+            # Check that one of the input UTxOs was spent to verify the TX was
+            # successfully submitted to the chain.
+            # An input is spent when its combination of hash and ix is not found in the list
+            # of current UTxOs.
+            # TODO: check that the transaction is 1-block deep (can't be done in CLI alone)
+            txin = txins[0]
+            txin_hash = txin.utxo_hash
+            txin_ix = txin.utxo_ix
+            utxo_data = self.get_utxo(txin.address)
+            is_spent = False
+            for u in utxo_data:
+                if txin_hash == u.utxo_hash and txin_ix == u.utxo_ix:
+                    # input was found
+                    break
+            else:
+                # input was not found
+                is_spent = True
+            if is_spent:
+                break
+        else:
+            raise CLIError(f"Transaction '{txid}' didn't make it to the chain (from '{tx_file}').")
+
     def send_tx(
         self,
         src_address: str,
@@ -2004,16 +2069,16 @@ class ClusterLib:
         deposit: Optional[int] = None,
         invalid_hereafter: Optional[int] = None,
         invalid_before: Optional[int] = None,
-        script_files: OptionalFiles = (),
         join_txouts: bool = True,
+        verify_tx: bool = True,
         destination_dir: FileType = ".",
     ) -> TxRawOutput:
-        """Build, Sign and Send a transaction.
+        """Build, Sign and Submit a transaction.
 
         Args:
             src_address: An address used for fee and inputs (if inputs not specified by `txins`).
             tx_name: A name of the transaction.
-            txins: An interable of `UTXOData`, specifying input UTxOs (optional).
+            txins: An iterable of `UTXOData`, specifying input UTxOs (optional).
             txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
             tx_files: A `TxFiles` tuple containing files needed for the transaction (optional).
             fee: A fee amount (optional).
@@ -2023,9 +2088,10 @@ class ClusterLib:
             deposit: A deposit amount needed by the transaction (optional).
             invalid_hereafter: A last block when the transaction is still valid (optional).
             invalid_before: A first block when the transaction is valid (optional).
-            script_files: A list of paths to script files (optional).
             join_txouts: A bool indicating whether to aggregate transaction outputs
                 by payment address (True by default).
+            verify_tx: A bool indicating whether to verify the transaction made it to chain
+                and resubmit the transaction if not (True by default).
             destination_dir: A path to directory for storing artifacts (optional).
 
         Returns:
@@ -2036,9 +2102,10 @@ class ClusterLib:
 
         if fee is None:
             witness_count_add = 0
-            if script_files:
+            if tx_files and tx_files.script_files:
                 # TODO: workaround for https://github.com/input-output-hk/cardano-node/issues/1892
-                witness_count_add += 5 * len(script_files)
+                all_scripts = list(itertools.chain.from_iterable(tx_files.script_files))
+                witness_count_add += 5 * len(all_scripts)
             fee = self.calculate_tx_fee(
                 src_address=src_address,
                 tx_name=tx_name,
@@ -2072,10 +2139,12 @@ class ClusterLib:
             tx_body_file=tx_raw_output.out_file,
             tx_name=tx_name,
             signing_key_files=tx_files.signing_key_files,
-            script_files=script_files,
             destination_dir=destination_dir,
         )
-        self.submit_tx(tx_signed_file)
+        if verify_tx:
+            self.submit_tx(tx_file=tx_signed_file, txins=tx_raw_output.txins)
+        else:
+            self.submit_tx_bare(tx_file=tx_signed_file)
 
         return tx_raw_output
 
@@ -2234,6 +2303,7 @@ class ClusterLib:
         fee: Optional[int] = None,
         ttl: Optional[int] = None,
         deposit: Optional[int] = None,
+        verify_tx: bool = True,
         destination_dir: FileType = ".",
     ) -> TxRawOutput:
         """Send funds - convenience function for `send_tx`.
@@ -2247,6 +2317,8 @@ class ClusterLib:
             ttl: A last block when the transaction is still valid
                 (deprecated in favor of `invalid_hereafter`, optional).
             deposit: A deposit amount needed by the transaction (optional).
+            verify_tx: A bool indicating whether to verify the transaction made it to chain
+                and resubmit the transaction if not (True by default).
             destination_dir: A path to directory for storing artifacts (optional).
 
         Returns:
@@ -2261,6 +2333,7 @@ class ClusterLib:
             fee=fee,
             deposit=deposit,
             destination_dir=destination_dir,
+            verify_tx=verify_tx,
         )
 
     def wait_for_new_block(self, new_blocks: int = 1) -> None:
